/*
 *  Rdweb : Multiple user request control R backend daemon.
 *  Copyright (C) 2005,06  ISM.AC.JP.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <netdb.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <pwd.h>
#include <sys/stat.h>
#include <dirent.h>
#include <ctype.h>
#include <signal.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <limits.h>
#include <strings.h>

#include "Rdaemon_sha1.h"

#include "Rdaemon_auth.h"

#define BUF_SIZE          8192
#define LOG_FILE_FMT      "/tmp/Rdaemon%04d%02d.log"
#define LISTENER_NUM      10
#define DEFAULT_PORT      10024
#define SID_MAX           128
#define EXPIRE_TIME       (60*60) /* 60min */
#define USERNAME_SIZE     128
#define PASSWORD_SIZE     128

/*
 *================== COMMON and IPC ================
 */
static key_t pid_key=0;
static pid_t*kill_pid=NULL;
static int   pid_shmid=0;

static int   ppid=0;
static int   semid=0;
/*
union semun {
      int              val;
      struct semid_ds *buf;
      unsigned short  *array;
#ifdef __linux
      struct seminfo  *__buf;
#endif
};
*/

static void Rdaemon_shm_lock()
{
   struct sembuf sb;
   sb.sem_num=0;
   sb.sem_op=-1;
   sb.sem_flg=0;
   if(-1==semop(semid,&sb,1))
      perror("shmop lock");
}
static void Rdaemon_shm_unlock()
{
   struct sembuf sb;
   sb.sem_num=0;
   sb.sem_op=1;
   sb.sem_flg=0;
   if(-1==semop(semid,&sb,1))
      perror("shmop unlock");
}

static int shmid = 0;

typedef struct SID_st{
      time_t        last_time;
      unsigned char sha1[SHA1_LENGTH];
      char          username[USERNAME_SIZE];
} SID_t;
static SID_t *SID=(SID_t *)NULL;


/* ============================= TOOL ========================= */
/*
 * Loging file.
 */
void loging(char *fmt, ...){
   FILE *logfd;
   struct tm tm_t;
   va_list ap;
   time_t now;
   char logfilename[256];
   
   int mask;

   now = time(NULL);
   tm_t = *localtime(&now);

   mask=umask(0);
   sprintf(logfilename,
	   LOG_FILE_FMT,
	   tm_t.tm_year + 1900,
	   tm_t.tm_mon  + 1);
   
   if(NULL == (logfd = fopen(logfilename,"a+"))) return;
   lockf(fileno(logfd), F_LOCK, 0);
   
   fprintf(logfd,
	   "%04d-%02d-%02d %02d:%02d:%02d ",
	   tm_t.tm_year + 1900,
	   tm_t.tm_mon  + 1,
	   tm_t.tm_mday,
	   tm_t.tm_hour,
	   tm_t.tm_min,
	   tm_t.tm_sec);
   
   va_start(ap,fmt);
   vfprintf(logfd, fmt, ap);
   va_end(ap);
   
   lockf(fileno(logfd), F_ULOCK, 0);
   fclose(logfd);
   umask(mask);
}


/*
 * argument split
 */
int args_split(char *buf, char **args, size_t limit)
{
   int   argc=0;
   int   esc=0;
    
   memset(args,0,sizeof(char *)*limit);

   while(args[argc]==NULL||*buf){
      if(!esc && (*buf == ' '||*buf=='\t')){
	 if(argc > limit){
	    return(argc);
	 }
	 *buf=0;
	 argc++;
	 for(buf++;*buf && (*buf==' '||*buf=='\t'); buf++);
	 continue;
      }
      else if (!esc && *buf == '"'){
	 esc=1;
      }
      else if (!esc && *buf == 0x27){
	 esc=2;
      }
      else if (esc==1 && *buf == '"'){
	 esc=0;
      }
      else if (esc==2 && *buf == 0x27){
	 esc=0;
      }
      if ( args[argc] == NULL ){
	 args[argc]=buf;
      }
      buf++;
   }
   argc++;
   return(argc);
}

/*
 * line buffer summary.
 */
const char *summary_line(char *buf){
   static char temp[BUF_SIZE];
   int i;
    
   memset(temp, 0, sizeof(temp));
   for(i=0; *buf && i<80 && i < sizeof(temp) ; i++){
      if(*buf=='\r'){
	 temp[i++]='\\';
	 temp[i]='r';
	 buf++;
      }
      else if(*buf=='\n'){
	 temp[i++]='\\';
	 temp[i]='n';
	 buf++;
      }
      else
	 temp[i]=*buf++;
   }
   return(temp);
}


void trim_crlf_buf(char *buf)
{
   int i;
   for(i=0; buf[i] && i < BUF_SIZE ; i++){
      switch( buf[i]){
	 case '\n':
	 case '\r':
	    buf[i]=0;
	    ;;
      }
   }
}

/* =============================== COMMAND ==================== */
/*
 * Command parse.
 */
const char *commandline_parse(char *buf){
   static char temp[BUF_SIZE];
   int i;

   memset(temp, 0, sizeof(temp));
   for(i=0; buf[i] > ' ' && i < sizeof(temp) ; i++)
      temp[i] = toupper(buf[i]);
   return(temp);
}

/*
 * Command parse argument.
 */
const char *commandline_parse_arg(char *buf){
   static char temp[BUF_SIZE];
   int i=0;
    
   memset(temp, 0, sizeof(temp));
   for(i=0; buf[i] != ' ' && i < sizeof(temp) ; i++);
   for(; buf[i]!=0 && isspace(buf[i]) && i < sizeof(temp) ; i++);
   strncpy(temp, buf+i, sizeof(temp));
   return(temp);
}


/*
 * ========================== command control. ====================
 */
void command_user(char *username, char *password, char *home, uid_t *uid, gid_t *gid, FILE *in, FILE *out, char *line)
{
   struct passwd *pwd_t;
   char   check_username[USERNAME_SIZE];
   
   strncpy(check_username,commandline_parse_arg(line),sizeof(check_username));
   trim_crlf_buf(check_username);
   setpwent();
   while(NULL!=(pwd_t = getpwent()) && strcmp(pwd_t->pw_name, check_username));
   endpwent();
   if( pwd_t != NULL && pwd_t->pw_uid >= 100 ){
      loging("[pid=%d] user [%s]\n",getpid(),check_username);	
      fprintf(out,"USER=%s\n",check_username);
      *gid=pwd_t->pw_gid;
      *uid=pwd_t->pw_uid;
      strncpy(home,pwd_t->pw_dir, PATH_MAX);
      strncpy(username, check_username, USERNAME_SIZE);
      strncpy(password, "", PASSWORD_SIZE);
   }else{
      loging("[pid=%d] can't found user [%s]\n",getpid(),username);	
      username[0] = 0;
      password[0] = 0;
      home[0]=0;
      *gid=*uid=0;
      fprintf(out,"%s",line);		    
   }
}

void command_pass(char *username, char *password, char *home, uid_t *uid, gid_t *gid, FILE *in, FILE *out, char *line)
{
   strncpy(password,commandline_parse_arg(line), PASSWORD_SIZE);
   trim_crlf_buf(password);
   if ( 0 == rdweb_pass(username,password)){
      int i,j;
      char sha1str[SHA1_LENGTH*2+1];
      Rdaemon_shm_lock();
      for(i=0;i<SID_MAX && 0 != strcmp(username,SID[i].username);i++); /* found name */
      if (!(i < SID_MAX)) for(i=0;i<SID_MAX && 0 != SID[i].username[0];i++); /* found free line */
      if (i >= SID_MAX){ /* expire with the thing which is the oldest in a hopeless thing */
	 time_t min_time = SID[0].last_time;
	 for(i=0 ; i<SID_MAX ; i++) if ( min_time > SID[i].last_time );
	 for(i=0 ; i<SID_MAX && SID[i].last_time != min_time ; i++) ;
      }
      if (i < SID_MAX){
         char buf[BUF_SIZE];
	 struct sha1_ctx sha_state;

	 strncpy(SID[i].username, username, USERNAME_SIZE);
	 SID[i].username[USERNAME_SIZE-1]=0;
	 SID[i].last_time = time(0);
	 sprintf(buf, "%lu:", SID[i].last_time);
	 strncat(buf, username, BUF_SIZE);
	 strncat(buf, password, BUF_SIZE);

	 sha1_init_ctx     (&sha_state);
	 sha1_process_bytes(buf, strlen(buf), &sha_state);
	 sha1_finish_ctx   (&sha_state, SID[i].sha1);

	 for(j=0 ; j<SHA1_LENGTH ; j++)
	    sprintf(sha1str+j*2, "%02x", SID[i].sha1[j] );
      }
      Rdaemon_shm_unlock();
      loging("[pid=%d] user [%s] pass [********]\n",getpid(),username);	
      if(setgid(*gid)){
	 loging("[pid=%d] can't change group id [%s:%d]\n",
		getpid(), username, *gid);
	 fprintf(out,"%s",line);
	 return;
      }
      if(setuid(*uid)){
	 loging("[pid=%d] can't change user id [%s:%d]\n",
		getpid(), username, *uid);
	 fprintf(out,"%s",line);
	 return;
      }
      if(chdir(home)){
	 loging("[pid=%d] can't chdir [%s:%s]\n",
		getpid(), username, home);
	 fprintf(out,"%s",line);		    
	 return;
      }
      setenv("USER", username, 1);
      setenv("LOGNAME", username, 1);
      setenv("HOME", home, 1);
      fprintf(out,"PASS=%s", sha1str);
      fprintf(out," ; ");
      fprintf(out,"HOME=%s\n", home);
   }else{
      loging("[pid=%d] can't found user [%s]\n",getpid(),username);	
      password[0] = 0;
      /*
	username[0] = 0;
	home[0] = 0;
	*gid=*uid=0;
	*/
      fprintf(out,"%s",line);		    
   }
}

#define a2h(x) ((unsigned)(('0'<=x&&x<='9')?(x-'0'):(('a'<=x && x<='f')?(x-'a'+10):(('A'<=x&&x<='F')?(x-'A'+10):(0)))))
void command_sid(char *username, char *password, char *home, uid_t *uid, gid_t *gid, FILE *in, FILE *out, char *line)
{
   struct passwd *pwd_t;
   char           buf[BUF_SIZE];
   unsigned char  sha1[SHA1_LENGTH];
   int i;
   
   strncpy(buf,commandline_parse_arg(line),sizeof(buf));
   trim_crlf_buf(buf);

   for (i=0; i<SHA1_LENGTH ; i++)
      sha1[i]=(a2h(buf[i<<1])<<4) + a2h(buf[(i<<1)+1]);

   Rdaemon_shm_lock();
   for(i=0;i<SID_MAX && 0 != strncmp((char *)sha1, (char *)SID[i].sha1, SHA1_LENGTH);i++);
   

   if( i < SID_MAX ){
      loging("[pid=%d] found sid [%s]\n",getpid(),SID[i].username);
      setpwent();
      while(NULL!=(pwd_t = getpwent()) && strcmp(pwd_t->pw_name, SID[i].username));
      endpwent();
      if( pwd_t != NULL && pwd_t->pw_uid >= 100 ){
	 *gid=pwd_t->pw_gid;
	 *uid=pwd_t->pw_uid;
	 SID[i].last_time = time(0);   /* update */
	 strncpy(home,pwd_t->pw_dir, PATH_MAX);
	 strncpy(username, SID[i].username, USERNAME_SIZE);
	 strncpy(password, "SID", PASSWORD_SIZE);
	 loging("[pid=%d] siduser [%s]\n",getpid(),username);
	 fprintf(out,"SID=%s ; HOME=%s\n", username, home );
	 Rdaemon_shm_unlock();
         loging("[pid=%d] user [%s] pass [********]\n",getpid(),username);	
	 if(setgid(*gid)){
	    loging("[pid=%d] can't change group id [%s:%d]\n",
		   getpid(), username, *gid);
	    fprintf(out,"%s",line);
	    return;
	 }
	 if(setuid(*uid)){
	    loging("[pid=%d] can't change user id [%s:%d]\n",
		   getpid(), username, *uid);
	    fprintf(out,"%s",line);
	    return;
	 }
	 if(chdir(home)){
	    loging("[pid=%d] can't chdir [%s:%s]\n",
		   getpid(), username, home);
	    fprintf(out,"%s",line);		    
	    return;
	 }
	 setenv("USER", username, 1);
	 return;
      }
   }
   Rdaemon_shm_unlock();
   loging("[pid=%d] can't found sid [%s]\n",getpid(),buf);	
   username[0] = 0;
   password[0] = 0;
   home[0]=0;
   *gid=*uid=0;
   fprintf(out,"%s",line);		    
}

static char basepath[PATH_MAX];
void command_base(char *line, FILE *in, FILE *out, char *username)
{
   fprintf(out,"BASE=%s\n",basepath);
}

void command_exec(char *line, FILE *in, FILE *out, char *username)
{
   char path[BUF_SIZE];
   char *args[1024];
   int  argc;
   int  pid;
   int  pipepico[2];
   int  pipepoci[2];
   FILE *fd;
   char buf[BUF_SIZE];
   strncpy(path, commandline_parse_arg(line), sizeof(path));
   trim_crlf_buf(path);
   argc=args_split(path,args,sizeof(args)/sizeof(char*));
   pipe(pipepico);
   pipe(pipepoci);

   loging("EXEC (USER=%s) %s %s\n", username,args[0],args[1]);
   switch (pid = fork()){
      case -1: /* fail */
	 loging("EXEC fork faild. (USER=%s)\n", username);
	 close(pipepico[0]);close(pipepico[1]);
	 close(pipepoci[0]);close(pipepoci[1]);
	 break;
      case 0: /* child */
	 close(pipepico[0]); /* parent side pipe close */
	 close(pipepoci[1]);
	 close(fileno(stdin));
	 close(fileno(stdout));

	 dup(pipepoci[0]);  /* pipe input */
	 dup(pipepico[1]);  /* pipe output */
	 execvp(args[0],args);
	 exit(0);
	 break;
      default:/* parent */
	 close(pipepico[1]); /* child side pipe close */
	 close(pipepoci[0]);
	 
	 close(pipepoci[1]); /* child <- EOF */

	 fd=fdopen(pipepico[0],"r");
	 while(fgets(buf,sizeof(buf),fd)){
	    fprintf(out,"%s",buf);
	 }
	 fprintf(out,"EOF\n");
	 fclose(fd);
   }
}
void command_getenv(char *line, FILE *in, FILE *out, char *username)
{
   char name[BUF_SIZE];
   strncpy(name, commandline_parse_arg(line), sizeof(name));
   trim_crlf_buf(name);
   fprintf(out,"GETENV=%s %s\n",name,getenv(name));
}

void command_setenv(char *line, FILE *in, FILE *out, char *username)
{
   char name[BUF_SIZE];
   char value[BUF_SIZE];
   int i;

   strncpy(name, commandline_parse_arg(line), sizeof(name));
   strncpy(value, commandline_parse_arg(name), sizeof(value));
   trim_crlf_buf(name);
   trim_crlf_buf(value);
   for(i=0;i<BUF_SIZE&&name[i];i++){
      if (name[i]==' ' ){
	 name[i]=0;
	 break;
      }
   }
   setenv(name, value, 1);
   fprintf(out,"SETENV=%s %s\n",name,value);
}

void command_pwd(char *line, FILE *in, FILE *out, char *username)
{
   char path[BUF_SIZE];

   getcwd(path, sizeof(path));
   fprintf(out,"PWD=%s\n",path);
}

void command_chdir(char *line, FILE *in, FILE *out, char *username)
{
   char path[BUF_SIZE];

   strncpy(path, commandline_parse_arg(line), sizeof(path));
   trim_crlf_buf(path);
   if(-1 == chdir( path )){
      loging("CHDIR faild. (USER=%s,chdir=%s)\n",
	       username,path);
      fprintf(out,"%s",line );
   }else{
      fprintf(out,"CHDIR=%s\n",path);
   }
}


void command_mkdir(char *line, FILE *in, FILE *out, char *username)
{
   char path[BUF_SIZE];

   strncpy(path, commandline_parse_arg(line), sizeof(path));
   trim_crlf_buf(path);
   if(-1 == mkdir( path, 0755 )){
      loging("MKDIR faild. (USER=%s,mkdir=%s)\n",
	     username,path);
      fprintf(out,"%s",line );
   }else{
      fprintf(out,"MKDIR=%s\n",path);
   }
}

#define BASENAME_FMT_LENGTH (strlen("CCYYMMDD_hhmmss"))
void command_rmdir(char *line, FILE *in, FILE *out, char *username)
{
   char path[BUF_SIZE];
   DIR *dird;
   struct dirent *dirent_t;

   strncpy(path, commandline_parse_arg(line), sizeof(path));
   trim_crlf_buf(path);

   if (0==strlen(path)){
      loging("RMDIR faild. (USER=%s,path=%s)\n",
	     username,path);
   }else if(NULL==(dird=opendir(path))){
      loging("RMDIR faild. (USER=%s,path=%s)\n",
	     username,path);
   }else{
      while(NULL!=(dirent_t = readdir(dird))){
	 struct stat st_t;
	 char fullpath[BUF_SIZE];
	 int  point = strlen(path) - BASENAME_FMT_LENGTH;
	 if ( point >=0 ){
	    strncpy(fullpath,path,sizeof(fullpath));
	    strncat(fullpath,"/", sizeof(fullpath));
	    strncat(fullpath,dirent_t->d_name,sizeof(fullpath));
	    if (!strncmp( path + point,
			  dirent_t->d_name ,
			  BASENAME_FMT_LENGTH)){
	       if(!stat(fullpath,&st_t) && S_ISREG(st_t.st_mode)){
		  if(-1 == unlink( fullpath )){
		     loging("RMDIR faild. (USER=%s,unkink=%s)\n",
			    username,fullpath);
		  }
	       }
	    }
	 }
      }
      closedir(dird);
   }

   if(-1 == rmdir( path )){
      loging("RMDIR faild. (USER=%s,rmdir=%s)\n",
	     username,path);
      fprintf(out,"%s",line );
   }else{
      fprintf(out,"RMDIR=%s\n",path);
   }
}

void command_unlink(char *line, FILE *in, FILE *out, char *username)
{
   char path[BUF_SIZE];

   strncpy(path, commandline_parse_arg(line), sizeof(path));
   trim_crlf_buf(path);
   if(-1 == unlink( path )){
      loging("UNLINK faild. (USER=%s,unkink=%s)\n",
	     username,path);
      fprintf(out,"%s",line );
   }else{
      fprintf(out,"UNLINK=%s\n",path);
   }
}

void command_ls(char *line, FILE *in, FILE *out, char *username)
{
   char path[BUF_SIZE];
   DIR *dird;
   struct dirent *dirent_t;
   strncpy(path, commandline_parse_arg(line), sizeof(path));
   trim_crlf_buf(path);
   if (0==strlen(path))
      getcwd(path, sizeof(path));
   
   if(NULL==(dird=opendir(path))){
      loging("LIST faild. (USER=%s,path=%s)\n",
	     username,path);
   }else{
      while(NULL!=(dirent_t = readdir(dird))){
	 struct stat st_t;
	 struct tm   tm_t;
	 char *p="f";
	 char fullpath[BUF_SIZE];
	 unsigned long len=0;
	 unsigned long ymd=0;
	 unsigned long hms=0;
	 strncpy(fullpath,path,sizeof(fullpath));
	 strncat(fullpath,"/", sizeof(fullpath));
	 strncat(fullpath,dirent_t->d_name,sizeof(fullpath));
	 if(0==stat(fullpath,&st_t)){
	    len = st_t.st_size;
	    tm_t = *localtime(&st_t.st_mtime);
	    
	    ymd =
	       (tm_t.tm_year + 1900) * 10000 +
	       (tm_t.tm_mon  + 1)    *   100 +
	       tm_t.tm_mday;
	    hms =
	       tm_t.tm_hour * 10000 +
	       tm_t.tm_min  *   100 +
	       tm_t.tm_sec ;
	    
	 }else{
	    fprintf(out,"#%s:%ld\n", fullpath,sizeof(fullpath));
	 }
	 if (S_ISFIFO(st_t.st_mode))      p = "p";
	 else if (S_ISCHR(st_t.st_mode))  p = "c";
	 else if (S_ISDIR(st_t.st_mode))  p = "d";
	 else if (S_ISBLK(st_t.st_mode))  p = "b";
	 else if (S_ISREG(st_t.st_mode))  p = "-";
	 else if (S_ISLNK(st_t.st_mode))  p = "l";
	 else if (S_ISSOCK(st_t.st_mode)) p = "s";
	 fprintf(out,"%s\t%10ld\t%08ld %06ld\t%s\n",
		 p,
		 len,
		 ymd,
		 hms,
		 dirent_t->d_name);
      }
      closedir(dird);
   }
   fprintf(out,"EOF\n");
}

void command_load(char *line, FILE *in, FILE *out, char *username)
{
   char path[BUF_SIZE];
   FILE *fd;
   unsigned char ibuf[256];
   unsigned char obuf[512]; /* ibuf * 2 */
   size_t len=0;
   struct stat st_t;
   strncpy(path, commandline_parse_arg(line), sizeof(path));
   trim_crlf_buf(path);
   if (0==strlen(path)){
      fprintf(out, "EOF\n");
      return;
   }
   /* check owner */
   if(-1 == stat(path, &st_t)){
      fprintf(out, "EOF\n");
      loging("[pid=%d] can't read file (%s:%s)\n",
	     getpid(),username,path);
      return;
   }
   if( st_t.st_uid != getuid ()){
      fprintf(out, "EOF\n");
      loging("[pid=%d] can't read other owner file (%s:%s)\n",
	     getpid(),username,path);
      return;
   }
   if(NULL == (fd = fopen(path, "r"))){
      fprintf(out, "EOF\n");
      loging("[pid=%d] can't read file (%s:%s)\n",
	     getpid(),username,path);
      return;
   }
   while(0==feof(fd)&&0==ferror(fd)){
      int i;
      len=fread(ibuf,1,sizeof(ibuf),fd);
      for( i = 0 ; i < len ; i++ ){
	 sprintf((char *)obuf+i*2,"%02x",ibuf[i]);
      }
      obuf[len*2] = 0;
      fprintf(out,"%s\n",obuf);
   }
   fclose(fd);
   fprintf(out, "EOF\n");
}


void command_save(char *line, FILE *in, FILE *out, char *username)
{
   char path[BUF_SIZE];
   FILE *fd;
   unsigned char ibuf[512]; /* obuf * 2 */
   unsigned char obuf[256];
   strncpy(path, commandline_parse_arg(line), sizeof(path));
   trim_crlf_buf(path);
   if (0==strlen(path)){
      loging("[pid=%d] user=%s can't filename (%s:%d)\n",
	     getpid(),username,path,strlen(path));
      fprintf(out, "EOF\n");
      return;
   }
   if(NULL == (fd = fopen(path, "w"))){
      fprintf(out, "EOF\n");
      loging("[pid=%d] can't write file (%s:%s)\n",
	     getpid(),username,path);
      return;
   }
   loging("[pid=%d] user=%s trans filename (%s)\n",
	   getpid(),username,path);
   /* tric 2 byte sets + 1 byte  cf ff01 -> 5 byte */
   while(fgets((char *)ibuf,((sizeof(ibuf)-1)/2)*2+1,in)){
      int i;
      size_t offset=0;
      size_t len=0;
      if(!strncmp("EOF", commandline_parse((char *)ibuf),3)){
	 break;
      }
      trim_crlf_buf((char *)ibuf);
      memset(obuf,0,sizeof(obuf));
      len = strlen((char *)ibuf);
      if ( len >= sizeof(ibuf) )
	 len = sizeof(ibuf);
      for( i = 0 ; i < len ; i++ ){
	 unsigned char u=0;
	 if (isdigit(ibuf[i])){
	    u = ibuf[i]-'0' ;
	 }else if  (isxdigit(ibuf[i])){
	    u = (tolower(ibuf[i]) - 'a') + 0xa;
	 }
	 offset+=i;
	 if(0==i%2)
	    u = u << 4;
	 obuf[i/2]+=u;
	 /*
	   fprintf(out,"%ld:%d:%d:%x:=>%x:\n",
	   strlen((char*)ibuf),i,i%2,u,obuf[i/2]);
	 */
      }
      fwrite(obuf, len/2, 1, fd);
   }
   fclose(fd);
   fprintf(out, "EOF\n");
}

void command_whodo(char *line, FILE *in, FILE *out, char *username)
{
   int  i;

   Rdaemon_shm_lock();
   for( i=0; i<SID_MAX; i++ ){
      if (SID[i].username[0]!=0){
	 fprintf ( out, "%u:%s:",
		   (unsigned int)SID[i].last_time,
		   SID[i].username);
	 fprintf( out, "\n");
      }
   }
   Rdaemon_shm_unlock();
   fprintf(out, "EOF\n");
}

typedef struct {
      char *comm_name;
      void (*comm_func)(char *, FILE *, FILE *, char *);
} comm_control_t;

static const comm_control_t comm_control[] = {
   { "BASE",    command_base   },
   { "EXEC",    command_exec   },
   { "GETENV",  command_getenv },
   { "SETENV",  command_setenv },
   { "PWD",     command_pwd    },
   { "CD",      command_chdir  },
   { "MKDIR",   command_mkdir  },
   { "RMDIR",   command_rmdir  },
   { "UNLINK",  command_unlink },
   { "LIST",    command_ls     },
   { "LOAD",    command_load   },
   { "SAVE",    command_save   },
   { "WHODO",   command_whodo  },
   { NULL,      NULL           },
};


/*
 * Cleanup Zombie.
 */
void cleanup_zombie(int sig)
{
   pid_t pid ;
   while( 0 < (pid=waitpid(-1,NULL,WNOHANG))) {
      loging("[pid=%d] zombie %d signal %d deleted.\n",getpid(),pid,sig);
      continue;
   }
   signal( SIGCHLD, cleanup_zombie);
}

/*
 * informatino host port.
 */
void host_port_info( int port_no, struct in_addr inaddr_no )
{
   char hostname[256] ;
   gethostname( hostname,sizeof(hostname) );
   hostname[sizeof(hostname)-1] = 0 ;
   loging("[pid=%d] run Rdaemon hostname=%s listner=%d \n",getpid(),hostname, port_no );
}

/*
 * file descpritoer dup socket.
 */
int fddup_sock( int sock, FILE **fdip, FILE **fdop )
{
   int sock_dup ;
   if( (sock_dup = dup(sock)) < 0 ) {
      return( -1 );
   }
   if( (*fdip = fdopen( sock_dup, "r" )) == NULL ) {
      close( sock_dup );
      return( -1 );
   }
   if( (*fdop = fdopen( sock, "w" )) == NULL ) {
      fclose( *fdip );
      *fdip = 0 ;
      return( -1 );
   }
   /* buffer control */
   setvbuf(*fdop, (char *)NULL, _IOLBF, 0);
   return( 0 );
}


/* ====================== CONTROL ===================== */

/*
 * command parsed Client control.
 */
void	Rdaemon_task( int comm )
{
   char  line[BUF_SIZE] ;
   const char *tmp;
   int   recv_size ;
   FILE *in, *out ;
   char  username[USERNAME_SIZE];
   char  password[PASSWORD_SIZE];
   char  home[PATH_MAX];
   uid_t uid=0;
   gid_t gid=0;
   

   memset(username, 0, sizeof(username));
   memset(password, 0, sizeof(password));
   memset(home, 0, sizeof(home));

   if( fddup_sock(comm,&in,&out) < 0 ) {
      loging("[pid=%d] fdooen() : %s\n", getpid(),strerror(errno));
      exit( 1 );
   }
   while( fgets(line,sizeof(line),in) ) {
      recv_size = strlen( line );
      
      /*
       * task select
       */
      if (!strcmp("QUIT", commandline_parse(line)))
	 break;
      
      if (!strcmp(password,"") && !strcmp("USER", commandline_parse(line))){
	 command_user(username, password, home, &uid, &gid, in, out, line);
	 continue;
      }
      if (strcmp(username,"") && !strcmp(password,"") && !strcmp("PASS", commandline_parse(line))){
	 command_pass(username, password, home, &uid, &gid, in, out, line);
	 continue;
      }
      if (!strcmp(password,"") && !strcmp("SID", commandline_parse(line))){
	 command_sid(username, password, home, &uid, &gid, in, out, line);
	 continue;
      }
      
      /* after user set. */
      if ( uid && strcmp(password,"")){
	 int idx;
	 for (idx = 0 ;
	      comm_control[idx].comm_name != NULL &&
		 strcmp(comm_control[idx].comm_name, commandline_parse(line)) ; idx++ );
	 if ( comm_control[idx].comm_name ){
	    comm_control[idx].comm_func(line, in, out, username);
	    continue;
	 }
      }
      
      tmp = summary_line(line);
      loging("[pid=%d] received (comm=%d) %d bytes, [%s]\n", getpid(), comm,recv_size, tmp );
      fflush( stdout );
      fprintf(out,"%s",line );
   }
   loging("[pid=%d] connection (comm=%d) closed.\n",getpid(),comm );
   fclose( in );
   fclose( out );
}

/*
 * socket information.
 */
void sockaddr_info( struct sockaddr *addrp, socklen_t addr_len )
{
   char host[BUF_SIZE] ;
   char port[BUF_SIZE] ;
   if( 0 > getnameinfo(addrp, 
		       addr_len,
		       host,
		       sizeof(host),
		       port,
		       sizeof(port),
		       NI_NUMERICHOST|NI_NUMERICSERV ))
      return;
   loging("[pid=%d] connection from %s:%s\n", getpid(), host, port );
}

/*
 * TCP information.
 */
void tcp_addr_info( int comm )
{
   struct sockaddr_storage addr ;
   socklen_t addr_len ;
   addr_len = sizeof( addr );
   if( 0 > getpeername( comm, (struct sockaddr *)&addr, &addr_len )) {
      perror("tcp_addr_info");
      return;
   }
   loging("[pid=%d] connection (comm=%d)\n",getpid(),comm );
   sockaddr_info( (struct sockaddr *)&addr, addr_len );
}

/*
 * TCP Accept port
 */
int tcp_listner_port( int port_no, struct in_addr inaddr_no )
{
   struct sockaddr_in addr ;
   int sock ;
   int val = 1;
   
   if( 0 > (sock = socket(PF_INET, SOCK_STREAM, 0))) {
      perror("socket");
      return( -1 );
   }
   
   setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(val));
   
   memset( &addr, 0, sizeof(addr) );
   addr.sin_family = AF_INET ;
   addr.sin_addr.s_addr = inaddr_no.s_addr ;
   /*
     addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
   */
   addr.sin_port = htons( port_no );
   
   if( bind(sock,(struct sockaddr *)&addr,sizeof(addr)) < 0 ){
      perror("bind");
      loging("port number %d is already used.\n", port_no );
      return( -1 );
   }
   if( listen( sock, LISTENER_NUM ) < 0 ) {
      perror("listen");
      close( sock );
      return( -1 );
   }
   return( sock );
}


static int exit_close(int rc)
{
   if ( pid_shmid > 0 && kill_pid !=NULL){
      shmdt((void *)kill_pid);
   }
   if ( shmid > 0 ) shmdt((void *)SID);
   if ( ppid == getpid()){      
      if(pid_shmid > 0 ) shmctl(shmid, IPC_RMID, NULL);
      if(semid > 0 )     semctl(semid,0,IPC_RMID,0);
      if(shmid > 0 )     shmctl(shmid, IPC_RMID, NULL);
   }
   return(rc);
}

static void exit_clean(int num)
{
   exit(exit_close(0));
}


void expire_SID(void)
{
   int i;
   Rdaemon_shm_lock();
   for( i=0 ; i<SID_MAX ; i++ )
   {
      if ( SID[i].username[0] != 0 && SID[i].last_time < time(0) - EXPIRE_TIME )
      {
	 loging("[pid=%d] expire [%s:%u]\n", getpid(), SID[i].username, SID[i].last_time );
	 SID[i].last_time = 0;
	 memset(SID[i].username, 0, USERNAME_SIZE);
	 memset(SID[i].sha1, 0, SHA1_LENGTH);
      }
   }
   Rdaemon_shm_unlock();
}

/*
 * Server Daemon.
 */
void Rdaemon_server( int port_no, struct in_addr inaddr_no )
{
   int listner,comm ;
   pid_t child_pid ;
    
   signal(SIGTERM,exit_clean);

   /* SHM PID */
   if ( -1 == ( pid_shmid = shmget( pid_key, sizeof(pid_t), IPC_CREAT|IPC_EXCL|S_IRUSR|S_IWUSR ))){
      perror("create pid on shmget");
      exit(exit_close(-1));
   }
   if ( (void*)-1 == ( kill_pid = shmat( pid_shmid, 0, 0))){
      perror("attach pid on shmat");
      exit(exit_close(-1));
   }
   *kill_pid = ppid;

   /* IPC SEM */
   if( -1 == ( semid = semget(IPC_PRIVATE,1,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH))){
      perror("semget");
      exit(exit_close(-1));
   }
   if( -1 == semctl(semid,0,SETVAL,1)){
      perror("semctl");
      exit(exit_close(-1));
   }

   /* IPC SHM */
   if ( -1 == ( shmid = shmget(IPC_PRIVATE, sizeof(SID_t)*SID_MAX, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH))){
      perror("shmget");
      exit(exit_close(-1));
   }
   if( (void *)-1 == ( SID = shmat(shmid, 0, 0 ))){
      perror("shmat");
      exit(exit_close(-1));
   }
   memset(SID,0,sizeof(SID_t)*SID_MAX);

   listner = tcp_listner_port( port_no, inaddr_no );
   if( 0 > listner )
      exit(exit_close(-1));
   host_port_info( port_no, inaddr_no );
   signal( SIGCHLD, cleanup_zombie );
   fprintf(stdout,"daemon process %u started.\n",getpid());
   for(;;) {
      expire_SID();
      errno=0;
      if( (comm = accept( listner,0,0 )) < 0 ) {
	 if ( errno == EINTR ) /* AIX etc...: delay interupt */
	    continue;
	 perror("accept");
	 exit(exit_close(-1));
      }
      tcp_addr_info( comm );
      if( (child_pid=fork()) > 0 ) { /* parent */
	 close( comm );
      }
      else if( child_pid == 0 ) { /* child */
	 close( listner );
	 Rdaemon_task( comm );
	 exit( 0 );
      } else {
	 perror("fork");
	 exit(exit_close(-1));
      }  
   }
}



/*
 *  usage
 */
static void usage(void){
   fprintf(stdout,"Usage: Rdaemon [-p #] [-h #] start|stop\n");
   fprintf(stdout," -p #  : port number (# is a value:default %d)\n", DEFAULT_PORT);
   fprintf(stdout," -h #  : any, loopback or hostname (# is a value:default localhost.)\n");
   exit(1);
}

/*************************************************************
 *  main
 *************************************************************/
#include <locale.h>
int main( int argc, char *argv[], char *arge[] )
{
   int port_no = DEFAULT_PORT;
   struct in_addr inaddr_no;
   int pid;
   int ch;
   extern int optind,opterr;
   extern char *optarg;
   char *p;


   printf(argv[0]);
   pid_key = ftok( argv[0], 0 );
   if(*argv[0]=='/')
      strncpy(basepath,argv[0],sizeof(basepath));
   else {
      getcwd(basepath, sizeof(basepath));
      strncat(basepath, "/", sizeof(basepath));
      strncat(basepath, argv[0], sizeof(basepath));
   }
    if(NULL != ( p = rindex( basepath, '/'))) *p='\0';
    
    /* inaddr_no.s_addr=htonl(INADDR_LOOPBACK); */
    inaddr_no.s_addr=htonl(INADDR_ANY);
    
    if ( getenv("DEBUG") == NULL ){
       if(-1 == setgid(0)){
	  perror("setgid");
	  exit(-1);
       }
       if(-1 == setuid(0)){
	  perror("setgid");
	  exit(-1);
       }
    }else
       fprintf(stderr,"running debug mode. uid=%d.\n",getuid());
    
    if( argc == 1 ){
       usage();
       exit(-1);
    }
    while((ch=getopt(argc,argv,"p:h:"))!=-1){
       switch(ch){
	  case 'p':
            port_no = strtol(optarg,NULL,10);
            break;
	  case 'h':
	     if ( optarg == NULL ){
		inaddr_no.s_addr=htonl(INADDR_ANY);
	     }else if (!strcmp(optarg,"any")){
		inaddr_no.s_addr=htonl(INADDR_ANY);
	     }else if (!strcmp(optarg,"loopback")){
		inaddr_no.s_addr=htonl(INADDR_LOOPBACK);
	     }else{
		inet_aton(optarg,&inaddr_no);
	     }
	     break;
	  default:
	     usage();
	     exit(-1);
       }
    }
    argc -= optind;
    argv += optind;
    
    if ( argc != 1 ){
       usage();
       exit(-1);
    }
    
    if (!strcmp(argv[0], "stop")){

       if ( -1 == ( pid_shmid = shmget( pid_key, sizeof(pid_t), IPC_EXCL|S_IRUSR|S_IWUSR ))){
	  exit(1);
       }
       if ( (void*)-1 == ( kill_pid = shmat( pid_shmid, 0, 0))){
	  shmctl(pid_shmid, IPC_RMID, NULL);
	  exit(2);
       }
       if(0==kill(*kill_pid,SIGTERM)){
	  fprintf(stdout,"daemon process %u stop.\n",*kill_pid);
       }
       shmdt((void *)kill_pid);
       shmctl(pid_shmid, IPC_RMID, NULL);

       exit(0);
    } else if ( strcmp(argv[0], "start")){
       usage();
       exit(-1);
    }
    
    signal(SIGHUP,SIG_IGN);
    
    switch(pid=fork()){
       case -1:
	  perror("fork");
	  exit(-1);
	  break;
       case 0:
	  ppid = getpid();
	  Rdaemon_server( port_no, inaddr_no );
	  break;
       default:
	  exit(0);
	  break;
    }
    printf("Why?\n");
    return(0);
}
