/*
 *  Rdweb : Multiple user request control R backend daemon.
 *  Copyright (C) 2005-07  ISM.AC.JP.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>

#if !defined(crypt)
extern char* crypt(const char *, const char *);
#endif

#define USERNAME_FIELD 1
#define PASSWD_FIELD 2

#ifndef    PAM_SERVICE_NAME
#define    PAM_SERVICE_NAME "other"
#endif  /* PAM_SERVICE_NAME */

#define getfield(_buf,_val,_sep,_fid) \
{\
   char *tok;\
   int   cid=0;\
   char *saveptr;\
\
   tok = strtok_r( _val, _sep, &saveptr );\
   while(tok != NULL){\
      cid++;\
      if ( cid == _fid ){\
	 strcpy(_buf,tok);\
	 break;\
      }\
      tok = strtok_r( NULL, _sep, &saveptr );\
   }\
}




/**********************************************************************
 * NIS                                                                *
 * ------------------------------------------------------------------ *
 **********************************************************************/
#ifdef USE_NIS
#include <rpc/rpc.h>
#include <rpc/xdr.h>
#include <rpcsvc/yp_prot.h>
#include <rpcsvc/ypclnt.h>
static int rdweb_auth_nis (
   char *username,
   char *password,
   char *domain               /* domain == NULL => nis default domain */
   )
{
   int   rc = -1;
   char *target_domain;

   int   result;
   char *val;
   int   val_len=0;
   int   username_len=strlen(username);

   char  pw_fld_buf[256];
   char *crypt_buf;
   char  salt[128];

   memset(pw_fld_buf,0,sizeof(pw_fld_buf));

   if ( domain == NULL ){
      if ((result = yp_get_default_domain(&target_domain)) != 0) {
	 return(rc);
      }
   }else{
      if ( NULL == ( target_domain = malloc(strlen(domain)+1))){
	 return(rc);
      }
      strcpy(target_domain,domain);
   }
      
   if ((result = yp_match(target_domain,"passwd.byname",
			  username, username_len,
			  &val, &val_len)) != 0){
      goto break_nis;
   }
      
   getfield(pw_fld_buf, val, ":\n", PASSWD_FIELD );
   free(val);
   rc=1;
      
   if(!strcmp("x",pw_fld_buf)){ /* shadow */
      if ((result = yp_match(target_domain,"shadow.byname",
			     username, username_len,
			     &val, &val_len)) != 0){
	 rc=0;
	 goto break_nis;
      }
	 
      getfield(pw_fld_buf, val, ":\n", PASSWD_FIELD );
      free(val);
      rc=1;
   }
  break_nis:
   free(target_domain);

   if (rc == 0) return -1;
   if ( 0 == strcmp( pw_fld_buf, "*")) return -1;
   if ( 0 == strcmp( pw_fld_buf, "!")) return -1;
   
   if ( 0 == strncmp( pw_fld_buf, "$1$", 3 )){ /* md5 */
      memcpy( salt, pw_fld_buf, sizeof(salt));
      
   } else { /* crypt */
      memcpy( salt, pw_fld_buf, 2 );
   }
   
   crypt_buf = crypt(password, salt);

   return(strcmp(crypt_buf,pw_fld_buf));
}
#endif /* USE_NIS */

/**********************************************************************
 * PAM                                                                *
 * ------------------------------------------------------------------ *
 **********************************************************************/
#ifdef USE_PAM
#include <pam_appl.h>
typedef struct {
    const char *username;
    const char *password;
    pam_handle_t *pamh;
} pam_appdata;

static int rdweb_auth_pam_conv (
   int msg_num,
   struct pam_message **msg,
   struct pam_response **resp,
  void *appdata_ptr
  )
{
   pam_appdata *l_appdata;
   struct pam_response *l_resp;
   int i;
   int y;
   const char *prompt;
   int rc;

   l_appdata = appdata_ptr;

   if(NULL == (l_resp = malloc(sizeof(struct pam_response) * msg_num)))
      return PAM_CONV_ERR;

   for (i = 0; i < msg_num ; i++)
   {
      switch (msg[i]->msg_style)
      {
         case PAM_PROMPT_ECHO_OFF:
            if( NULL == (l_resp[i].resp = strdup(l_appdata->password)))
               goto rdweb_pam_conv_error;
            l_resp[i].resp_retcode = PAM_SUCCESS;
            break;

         case PAM_PROMPT_ECHO_ON:
            if ( PAM_SUCCESS != ( rc = pam_get_item(l_appdata->pamh,
                                                    PAM_USER_PROMPT,
                                                    (void *) &prompt)))
               goto rdweb_pam_conv_error;

            if (!strcmp(msg[i]->msg, prompt))
               l_resp[i].resp = strdup(l_appdata->username);
            else
               l_resp[i].resp = NULL;
            l_resp[i].resp_retcode = PAM_SUCCESS;
            break;

         case PAM_ERROR_MSG:
         case PAM_TEXT_INFO:
            l_resp[i].resp = NULL;
            l_resp[i].resp_retcode = PAM_SUCCESS;
            break;
         default:
            goto rdweb_pam_conv_error;
      }
   }
   *resp = l_resp;
   return PAM_SUCCESS;

rdweb_pam_conv_error:

   for (y = 0; y < i; y++)
      if (l_resp[y].resp != NULL)
         free(l_resp[y].resp);
   free(l_resp);
   return PAM_CONV_ERR;
}

static int rdweb_auth_pam (
   char *username,
   char *password,
   char *service      /* service == NULL => default set SERVICE */
   )
{

   struct pam_conv l_conv;
   pam_appdata l_appdata;
   pam_handle_t *pamh;
   int rc;

   if ( service == NULL ) service = PAM_SERVICE_NAME; 

   l_appdata.username = username;
   l_appdata.password = password;
   l_appdata.pamh = NULL;

#if defined(__linux)
   l_conv.conv = (void *)rdweb_auth_pam_conv;
#else
   l_conv.conv = rdweb_auth_pam_conv;
#endif
   l_conv.appdata_ptr = &l_appdata;

   if ( PAM_SUCCESS != (rc = pam_start(service,
                                       username,
                                       &l_conv,
                                       &pamh)))
      return -1;

   l_appdata.pamh = pamh;

   if ( PAM_SUCCESS != (rc = pam_authenticate(pamh, PAM_SILENT))){
      pam_end(pamh, rc);
      return -1;
   }
   
   if ( PAM_SUCCESS != (rc = pam_acct_mgmt(pamh, PAM_SILENT))){
      pam_end(pamh, rc);
      return -1;
   }

   pam_end(pamh, PAM_SUCCESS);
   
   return 0;
}


#endif /* USE_PAM */


#if !(defined(USE_PAM)||defined(USE_NIS))
/**********************************************************************
 * FILE                                                               *
 * ------------------------------------------------------------------ *
 **********************************************************************/
static int rdweb_auth_file(char *username, char *password)
{
   char  pw_fld_buf[4096];
   char *crypt_buf;
   char  salt[12];

   char *etc_passwd = "/etc/passwd";
   char *etc_shadow = "/etc/shadow";
   FILE *fd;
   char  fbuf[4096];
   int   found=0;
   char user_fld_buf[4096];
   char savebuf[4096];

   memset(pw_fld_buf,0,sizeof(pw_fld_buf));


   if(NULL == (fd = fopen( etc_passwd, "r" ))) return -1;
   while (fgets(fbuf, sizeof(fbuf), fd)){
      strcpy(savebuf,fbuf);
      getfield(user_fld_buf, fbuf, ":\n", USERNAME_FIELD );
      if(0==strcmp(user_fld_buf, username)){
	 found=1;
	 getfield(pw_fld_buf, savebuf, ":\n", PASSWD_FIELD );
	 break;
      }
   }
   fclose(fd);

   if(found && 0==strcmp(pw_fld_buf,"x")){     
      if(NULL == (fd = fopen( etc_shadow, "r" ))) return -1;
      while (fgets(fbuf, sizeof(fbuf), fd)){
	 strcpy(savebuf,fbuf);
	 getfield(user_fld_buf, fbuf, ":\n", USERNAME_FIELD );
	 if(0==strcmp(user_fld_buf, username)){
	    getfield(pw_fld_buf, savebuf, ":\n", PASSWD_FIELD );
	    break;
	 }
      }
      fclose(fd);
   }


   if (found == 0) return -1;
   if ( 0 == strcmp( pw_fld_buf, "*")) return -1;
   if ( 0 == strcmp( pw_fld_buf, "!")) return -1;
   
   if ( 0 == strncmp( pw_fld_buf, "$1$", 3 )){ /* md5 */
      memcpy( salt, pw_fld_buf, sizeof(salt));
      
   } else { /* crypt */
      memcpy( salt, pw_fld_buf, 2 );
   }
   
   crypt_buf = (char *)crypt(password, salt);

   return(strcmp(crypt_buf,pw_fld_buf));
}
#endif /* !(defined(USE_PAM)||defined(USE_NIS)) */


int rdweb_pass(char* username, char* password)
{
#ifdef USE_PAM
   if(0 == rdweb_auth_pam(username,password,NULL))
      return(0);
#endif

#ifdef USE_NIS
   if(0 == rdweb_auth_nis(username,password,NULL))
      return(0);
#endif

#if !(defined(USE_PAM)||defined(USE_NIS))
   if(0 == rdweb_auth_file(username,password))
      return(0);
#endif
   return(-1);
}
