#!/usr/bin/perl
#
#  Rdweb : Multiple user request control R frontend.
#  Copyright (C) 2005,06,07  ISM.AC.JP.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
#



use CGI qw(:standard :form :html3);
use CGI::Carp qw(fatalsToBrowser set_message);
use Rdweb;
use strict;
use Cwd;
use POSIX;


use IO::Socket;

my $RdwebJS = "scripts/rdweb.js";
my $H1hostname =  "<span id=\"hostname\">(${Rdweb::host})</span>";
my $COPYRIGHT = "<div id=\"copyright\">";
open(FD,  "../COPYRIGHT");
while(<FD>){ $COPYRIGHT .= $_ ;}
close(FD);
$COPYRIGHT .= "</div>";

BEGIN {
    sub handle_errors {
        my $msg = shift;
        $msg =~ s/(at.*?)$//;
my $errhtml = <<"EOF";
<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" lang="en_US">
<head>
  <title>Rdweb</title>
  <link rel="stylesheet" type="text/css" href="index.${Rdweb::host}.css" />
</head>
<body>
  <h1>Let&#x27;s Use Parallel <img height=36 src="images/Rlogo-5.gif" alt="R" />.$H1hostname</h1>
  $COPYRIGHT
  <p>$msg</p>
</body>
</html>
EOF
         print $errhtml;
    }
    set_message(\&handle_errors);
}

my $EOL="\r\n";

sub socket_open(){
    my $sock = IO::Socket::INET->new( Proto     => "tcp",
				      PeerAddr  => $Rdweb::host,
				      PeerPort  => $Rdweb::port,
				      );
    unless ($sock) { die "can't connect to R daemon on $Rdweb::host :$Rdweb::port " }
    $sock->autoflush(1);
    return $sock;
}

sub cancelJOB($$){
    my $q = shift;
    my $getsid = shift;
    
    my $sock = socket_open();
    
    my $user = RemoteUSER($q,$sock,undef,undef,$getsid);
    if( $user eq "" ){
	close($sock);
	RdwebLogin($q);
	return;
    }  my $home = RemotePWD($sock);
    my $rdir = RemoteMKDIR($sock,"$home/$Rdweb::dir");
    $rdir = RemoteCD($sock,"$home/$Rdweb::dir");
    
    my $remote_basepath=RemoteBASE($sock);
    my $path = RemoteGETENV($sock,"PATH") .":". $remote_basepath . "/../bin";
    RemoteSETENV($sock,"PATH",$path);
    
    my $cancel_command = $Rdweb::job_comm{Cancel};
    $cancel_command =~ s/\$RD_JOBID/$q->param('do_cancel')/eg;
    $cancel_command =~ s/\$RD_USER/$user/eg;
    RemoteSETENV($sock,"RD_JOBID",$q->param('do_cancel'));
    RemoteSETENV($sock,"RD_USER",$user);
    
    print $sock "exec $cancel_command" . $EOL;
    print FD "exec $cancel_command" . $EOL;
    while ( my $buf = <$sock> ){
	last if ( $buf =~ /^EOF/ );
    }
    close ($sock);
}


sub submitJOB($$){
    my $q = shift;
    my $getsid = shift;
    
    my($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
    my $src = $q->param('source');
    my $srctest = $src;
    
    $srctest =~ s/[ \n\t]//g;
    return unless (length($srctest)>0);
    
    my $usequeue = $q->param('useQueue');
    my $useblas = $q->param('useBLAS');
    my $usenode = 1;
    my $useslave = $q->param('useSlave');
    my $cltype = $q->param('cltype');
    my $basename = sprintf ( "%04d%02d%02d_%02d%02d%02d",
			     $year+1900, $mon+1, $mday, $hour, $min, $sec);
    my $codename  = "${basename}.R";
    my $outname   = "${basename}.Rout";
    my $shellname = "${basename}.sh";
    
    my $sock = socket_open();
    
    my $user = RemoteUSER($q,$sock,undef,undef,$getsid);
    if( $user eq "" ){
	close($sock);
	RdwebLogin($q);
	return;
    }
    
    my $mydir = ${Rdweb::work} ."/". $user ;

    mkdir ( ${Rdweb::work} ) or die "mkdir ${Rdweb::work} $!" . ${Rdweb::work} unless ( -d ${Rdweb::work} );
    mkdir ( $mydir ) or die "mkdir $mydir $!" . $mydir unless ( -d $mydir );
    
    ### generate.

    ### make R script
    $src =~ s/\r\n/\n/g;
    $src =~ s/\r/\n/g;
    open(FD, "> ${mydir}/${codename}") or die "${mydir}/${codename} $!:Permission denied";
    printf FD "%s",$src;
    printf FD "\n";
    close(FD);
    
    ### make shell script
    my $shellscript = "";
    if ( $useslave eq "0" ) {
	open(FD, "../skel/$Rdweb::host/single.submit.sh") ;
    }else{
	if ( $cltype eq "MPI" ) {
	    open(FD, "../skel/$Rdweb::host/mpi.submit.sh") ;
	} else {
	    open(FD, "../skel/$Rdweb::host/sock.submit.sh") ;
	}
    }
    while(<FD>){
	$shellscript .= $_;
    }
    close(FD);

    my $caljobtype = "serial";
    $caljobtype = "parallel" unless ( $useslave eq "0" );
    my $caltasks = 0;
    my $calcpus = 0;
    if (  $caljobtype eq "serial" ){
        $caltasks =1;
	$calcpus  =1;
    }else{
        $caltasks += $Rdweb::manager_count;
        $calcpus += $Rdweb::manager_count;
    }
    $caltasks += $useslave if ( $useslave ne "0" );
    $calcpus += $useslave if ( $useslave ne "0" );
    $calcpus *= $useblas;
    
    $usenode = ceil($calcpus/$Rdweb::MAXNODECPU);
    
    $shellscript =~ s/\$RD_QUEUENAME/$usequeue/g;
    $shellscript =~ s/\$RD_JOBTYPE/$caljobtype/g;
    $shellscript =~ s/\$RD_NODE/$usenode/g;
    $shellscript =~ s/\$RD_TASKS/$caltasks/g;
    $shellscript =~ s/\$RD_CPUS/$calcpus/g;
    $shellscript =~ s/\$RD_SLAVE/$useslave/g;
    $shellscript =~ s/\$RD_BASENAME/$basename/g;
    $shellscript =~ s/\$RD_RSHCMD/$Rdweb::rshcmd/g;
    $shellscript =~ s/\$RD_MANAGER_COUNT/$Rdweb::manager_count/g;
    $shellscript =~ s/\$RD_BLASENV/$Rdweb::blasenv/g;
    $shellscript =~ s/\$RD_BLASNUM/$useblas/g;
    $shellscript =~ s/\$RD_CLTYPE/$cltype/g;
    

    open(FD, "> ${mydir}/${shellname}") or die "${mydir}/${shellname} $!: Permission denied";
    printf FD $shellscript;
    close(FD);

    ### trans.
    my $home = RemotePWD($sock);
    my $rdir = RemoteMKDIR($sock,"$home/$Rdweb::dir");
    my $rjobdir = RemoteMKDIR($sock,"$home/$Rdweb::dir/$basename");
    $rjobdir = RemoteCD($sock,"$home/$Rdweb::dir/$basename");
    
    my $pwd = getcwd;
    chdir(${mydir});
    RemoteSEND($sock, $shellname);
    RemoteSEND($sock, $codename);
    unlink($shellname);
    unlink($codename);
    chdir($pwd);
    
    $rdir = RemoteCD($sock,"$home/$Rdweb::dir");

    my $remote_basepath=RemoteBASE($sock);
    my $path = RemoteGETENV($sock,"PATH") .":". $remote_basepath . "/../bin";
    RemoteSETENV($sock,"PATH",$path);
    
    # R_PROFILE
    my $remote_basepath=RemoteBASE($sock);
    my $r_profile = $remote_basepath . "/../R/Rdweb.Rprofile"; 
    RemoteSETENV($sock,"R_PROFILE",$r_profile);
    
    # see. Rdweb.Rprofile
    RemoteSETENV($sock,"RD_QUEUENAME",$usequeue);
    RemoteSETENV($sock,"RD_JOBTYPE",$caljobtype);
    RemoteSETENV($sock,"RD_NODE",$usenode);
    RemoteSETENV($sock,"RD_TASKS",$caltasks);
    RemoteSETENV($sock,"RD_CPUS",$calcpus);
    RemoteSETENV($sock,"RD_SLAVE",$useslave);
    RemoteSETENV($sock,"RD_BASENAME",$basename);
    RemoteSETENV($sock,"RD_BLASNUM",$useblas);
    RemoteSETENV($sock,"RD_RSHCMD",$Rdweb::rshcmd);
    RemoteSETENV($sock,"RD_MANAGER_COUNT",$Rdweb::manager_count);
    RemoteSETENV($sock,$Rdweb::blasenv,$useblas);
    RemoteSETENV($sock,"RD_CLTYPE",$cltype);
    
    my $submit_command = $Rdweb::job_comm{Submit};
    $submit_command =~ s/\$RD_BASENAME/$basename/g;
    $submit_command =~ s/\$RD_QUEUENAME/$usequeue/g;
    $submit_command =~ s/\$RD_JOBTYPE/$caljobtype/g;
    $submit_command =~ s/\$RD_NODE/$usenode/g;
    $submit_command =~ s/\$RD_TASKS/$caltasks/g;
    $submit_command =~ s/\$RD_CPUS/$calcpus/g;
    $submit_command =~ s/\$RD_SLAVE/$useslave/g;
    $submit_command =~ s/\$RD_BASENAME/$basename/g;
    $submit_command =~ s/\$RD_RSHCMD/$Rdweb::rshcmd/g;
    $submit_command =~ s/\$RD_MANAGER_COUNT/$Rdweb::manager_count/g;
    $submit_command =~ s/\$RD_BLASENV/$Rdweb::blasenv/g;
    $submit_command =~ s/\$RD_BLASNUM/$useblas/g;
    $submit_command =~ s/\$RD_CLTYPE/$cltype/g;
    
    print $sock "exec $submit_command" . $EOL;

    while ( my $buf = <$sock> ){
	last if ( $buf =~ /^EOF/ );
    }

    close($sock);
}

sub RemoteSEND($$){
    my $sock = shift;
    my $file = shift;
    
    open (FD , $file);
    print $sock "save $file".$EOL;
    my $buf;
    my $length;
    while($length = sysread(FD, $buf,8192)){
	last unless ($length > 0);
	print $sock unpack("H*",$buf), $EOL;
    }
    print $sock "EOF".$EOL;
    <$sock>; # EOF
}

sub RemoteGETENV($$){
    my $sock = shift;
    my $name = shift;
    
    print $sock "getenv $name" . $EOL;
    my $buf = <$sock>;
    unless ( $buf =~ /^GETENV=/ ){
	close($sock);
	die "failed getenv $name.";
    }
    $buf =~ s/^GETENV=//g;
    $buf =~ s/^$name //g;
    $buf =~ s/[\r\n]//g;
    return($buf);
}

sub RemoteSETENV($$$){
    my $sock = shift;
    my $name = shift;
    my $value = shift;
    
    print $sock "setenv $name $value" . $EOL;
    my $buf = <$sock>;
    unless ( $buf =~ /^SETENV=/ ){
	close($sock);
	die "failed setenv $name $value.";
    }
    $buf =~ s/^SETENV=//g;
    $buf =~ s/^$name //g;
    $buf =~ s/[\r\n]//g;
    return($buf);
}


sub RemoteBASE($){
    my $sock = shift;
    
    print $sock "base" . $EOL;
    my $buf = <$sock>;
    unless ( $buf =~ /^BASE=/ ){
	close($sock);
	die "failed get base directory.";
    }
    $buf =~ s/^BASE=//g;
    $buf =~ s/[\r\n]//g;
    return($buf);
}

sub RemotePWD($){
    my $sock = shift;
    
    print $sock "pwd" . $EOL;
    my $buf = <$sock>;
    unless ( $buf =~ /^PWD=/ ){
	close($sock);
	die "failed get working directory.";
    }
    $buf =~ s/^PWD=//g;
    $buf =~ s/[\r\n]//g;
    return($buf);
}

sub RemoteMKDIR($$){
    my $sock = shift;
    my $dir = shift;
    
    my $cpwd = RemotePWD($sock);
    
    print $sock "cd $dir" . $EOL;
    my $buf = <$sock>;
    unless ( $buf =~ /^CHDIR=/ ){
	print $sock "mkdir $dir" . $EOL;
	my $buf = <$sock>;
	unless ( $buf =~ /^MKDIR=/ ){
	    close($sock);
	    die "failed make $dir directory.";
	}
    }else{
	print $sock "cd $cpwd" . $EOL;
	my $buf = <$sock>;
    }
    $buf =~ s/^CHDIR=//g;
    $buf =~ s/^MKDIR=//g;
    $buf =~ s/[\r\n]//g;
    return($buf);
}

sub RemoteCD($$){
    my $sock = shift;
    my $dir = shift;
    
    print $sock "cd $dir" . $EOL;
    my $buf = <$sock>;
    unless ( $buf =~ /^CHDIR=/ ){
	close($sock);
	die "failed change $dir $buf directory.";
    }
    $buf =~ s/^CHDIR=//g;
    $buf =~ s/[\r\n]//g;
    return($buf);
}

sub RemoteLIST($$){
    my $sock = shift;
    my $dir = shift;
    
    my @dirlist;
    
    print $sock "list $dir" . $EOL;
    while(my $buf = <$sock>){
	last if ( $buf =~ /^EOF/ );
	$buf =~ s/[\r\n]//g;
	my ($type, $size, $datetime, $filename) = split("\t", $buf);
	next if ( $filename =~ /^[\.]/ );
	push( @dirlist, $buf);
    }
    return(@dirlist);
}

sub RemoteUSER($$$$$){
    my $q    = shift;
    my $sock = shift;
    my $user = shift;
    my $pass = shift;
    my $sid  = shift;
    my $buf = "";

    if((defined($user)&&($user ne ""))&&(defined($pass)&&($pass ne ""))){
	print $sock "user " . $user . $EOL;
	$buf = <$sock>;
	unless ( $buf =~ /^USER=/ ){
	    close($sock);
	    RdwebLogin($q);
	    exit;
	}
	$buf =~ s/^USER=//g;
	$buf =~ s/[\r\n]//g;
	
	print $sock "pass " . $pass . $EOL;
	$buf = <$sock>;
	unless ( $buf =~ /^PASS=/ ){
	    close($sock);
	    RdwebLogin($q);
	    exit;
	}
	$buf =~ s/^PASS=//g;
	$buf =~ s/[\r\n]//g;
	$buf =~ s/\s;\sHOME=(.*)$//g;
    }
    if (defined($sid)&&($sid ne "")){
	print $sock "sid " . $sid . $EOL;
	$buf = <$sock>;
	unless ( $buf =~ /^SID=/ ){
	    close($sock);
	    return("");
	}
	$buf =~ s/^SID=//g;
	$buf =~ s/[\r\n]//g;  
	$buf =~ s/\s;\sHOME=(.*)$//g;
    }
    return($buf);
}

sub rep($$)
{
    my $var = shift;
    my $cnt = shift;
    my $ret = "";
    for ( my $i=0; $i<$cnt && $i< 10 ; $i++ ){
	$ret.=$var;
    }
    return($ret);
}

sub cutparm($$)
{
    my $line = shift;
    my $key  = shift;

    my $ret = $line;

    $ret =~ s/^(.*)$key([0-9]*)(.*)$/$2/e;

    return($ret);
}

sub getdirtree($$$$$)
{
    my $sock = shift;
    my $rdir = shift;
    my $ldir = shift;
    my $level = shift;
    my $mixdirlist = shift;

    my @dirlist = RemoteLIST($sock,$rdir);
    foreach my $dirent (@dirlist){
	my ($type, $size, $datetime, $filename ) = split("\t", $dirent);
	if ( $type eq 'd' ){
	    my $c_rdir = (($rdir eq "")?"":$rdir."/").$filename ;
	    my $c_ldir = $filename;
	    push(@$mixdirlist, join("\t", $type, $size, $level,$datetime, $c_rdir, $c_ldir));
	    getdirtree($sock, $c_rdir, $c_ldir, $level + 1, \@$mixdirlist);
	}else{
	    push(@$mixdirlist, join("\t", $type, $size, $level,$datetime, $rdir, $ldir, $filename));
	}
    }
}

sub Rdweb{
    my $q = shift;
    my $getsid = shift;
    
    my $buf;
    
    ## Daemon task ##
    my $sock = socket_open();
    my $user="";
    
    
    unless(defined($getsid)||defined($q->param('login'))){
	close($sock);
	RdwebLogin($q);
	return;
    }
    unless(defined($getsid)){
	$getsid = RemoteUSER($q,$sock,$q->param('Rdwebue'),$q->param('Rdwebpd'),undef);
	$user = $q->param('Rdwebue');
	if( $getsid eq "" ){
	    close($sock);
	    RdwebLogin($q);
	    return;
	}
    }else{
	$user = RemoteUSER($q,$sock,undef,undef,$getsid);
	if( $user eq "" ){
	    close($sock);
	    RdwebLogin($q);
	    return;
	}
    }
    
    my $home = RemotePWD($sock);
    my $rdir = RemoteMKDIR($sock,"$home/$Rdweb::dir");
    $rdir = RemoteCD($sock,"$home/$Rdweb::dir");


    my @mixdirlist=();
    getdirtree($sock,"","",0,\@mixdirlist);


    RemoteSETENV($sock,"RD_USER",$user);
    
    my $remote_basepath=RemoteBASE($sock);
    my $path = RemoteGETENV($sock,"PATH") .":". $remote_basepath . "/../bin";
    RemoteSETENV($sock,"PATH",$path);
    
    my $queue_list_command = $Rdweb::job_comm{QueueList};
    $queue_list_command =~ s/\$RD_USER/$user/eg;
    RemoteSETENV($sock,"RD_USER",$user);
    print $sock "exec $queue_list_command" . $EOL;
    my @jobs = ();
    while ( my $buf = <$sock> ){
	last if ( $buf =~ /^EOF/ );
	my %rec ;
	my @ary = split (' ', $buf );
	$rec{JOBID}=$ary[0];
	$rec{JOBQ}=$ary[1];
	$rec{JOBSTAT}=$ary[2];
	push (@jobs, \%rec);
    }
    close($sock);
    ## Daemon task ##
    
    my %skeltypecode=(""=>"[Select Sample Code]");
    opendir(DIR, "sample");
    my @fl = grep { !/^\./ && -f "sample/$_" } readdir(DIR);
    closedir DIR;
    foreach my $ex (@fl){
	my $data="";
	open(SFD, "sample/$ex");
	while(<SFD>){
	    $data.=$_;
	}
	close(SFD);
	$skeltypecode{$data}=$ex;
    }
    

    {
	my $fqdn = $q->url(-full=>1);
	my $secure = 0;
	if ( $fqdn =~ /https:\/\// ){
	    $secure=1;
	}
	$fqdn =~ s/^([0-9A-Za-z]+):\/\///;
	my $cookiepath = $q->url(-absolute=>1);
	$cookiepath =~ s/[^\/]+$//g;
	print $q->header(
			 -cookie=>$q->cookie(-name=>'SID',
					     -value=>$getsid,
					     -expires=>'+1h',
					     -path=>$cookiepath,
					     -host=>$fqdn,
					     -secure=>$secure
					     ),
			 -expires=>"now",
			 -charset=>'utf-8');
    }
    print $q->start_html(-title=>'Rdweb',
			 -script=>[{-src=>${Rdweb::checkjs}},
				   {-src=>$RdwebJS},
				   ],
			 -onLoad=>"source_update();rdweb_check_useQueue(document.Rdweb);",
			 -style=>{'src'=>"index.${Rdweb::host}.css"},
			 -lang=>'en_US',
			 -encoding=>'utf-8');
    
    print $q->h1({-Style=>"cursor: pointer",
		  -onClick=>
	              "do_action_clean();".
		      "document.Rdweb.submit();"
		  },
		 "Let's Use Parallel <img height=36 src=\"images/Rlogo-5.gif\" alt=\"R\" />.$H1hostname");
    print $COPYRIGHT;

    print ("Information is updated when you click a title area.");
    print $q->start_multipart_form(-name=>"Rdweb");
    $q->param(-name=>'MAXNODECPU',-value=>$Rdweb::MAXNODECPU);
    print $q->hidden(-name=>'MAXNODECPU');
    $q->param(-name=>'MANAGER_COUNT',-value=>$Rdweb::manager_count);
    print $q->hidden(-name=>'MANAGER_COUNT');
    $q->param(-name=>'sourcefontsize',-value=>"9pt") unless (defined($q->param('sourcefontsize')));
    print $q->hidden(-name=>'sourcefontsize');


    print $q->start_div({-id=>"main"});
    print $q->start_table;
    print $q->caption('Sequentially choose the setting concerning the job attribute from the left side.');
    print $q->Tr(
		 [
		  th(
		     [ "Queue","Slave","BLAS","Cluster Type"]
		     ),
		  td(
		     [
		      $q->popup_menu(-name=>'useQueue',
		                     -onChange=>'rdweb_check_useQueue(this.form);',
				     -values=>[sort {
					 cutparm(${Rdweb::Queue}{$a},"CPU=") <=> cutparm(${Rdweb::Queue}{$b},"CPU=") or
					     cutparm(${Rdweb::Queue}{$a},"MEM=") <=> cutparm(${Rdweb::Queue}{$b},"MEM=") or
					     cutparm(${Rdweb::Queue}{$a},"TIME=") <=> cutparm(${Rdweb::Queue}{$b},"TIME=") or
					     ${Rdweb::Queue}{$a} cmp ${Rdweb::Queue}{$b}
				     }
					       (keys(%Rdweb::Queue))],
				     -labels=>\%Rdweb::Queue),
		      $q->popup_menu(-name=>"useSlave",
		                     -onChange=>'rdweb_check_useSlave(this.form);',
				     -values=>[sort {$a <=> $b} (keys(%Rdweb::Slave))],
				     -labels=>\%Rdweb::Slave),
		      $q->popup_menu(-name=>"useBLAS",
		                     -onChange=>'rdweb_check_useBLAS(this.form);',
				     -values=>[sort {$a <=> $b} (keys(%Rdweb::BLAS))],
				     -labels=>\%Rdweb::BLAS),
		      $q->popup_menu(-name=>"cltype",
				     -values=>[sort {$a cmp $b} (keys(%Rdweb::Cltype))],
				     -labels=>\%Rdweb::Cltype,
				     -default=>$Rdweb::Cltype::default
				     ),
		      ]
		     )
		  ]
		 );
    print $q->end_table();


    my %sourcefsize = (8=>"8pt",
		       9=>"9pt",
		       10=>"10pt",
		       11=>"11pt",
		       12=>"12pt",
		       13=>"13pt",
		       14=>"14pt",
		       16=>"16pt");

    print $q->start_table;
    print $q->caption('Script Area');
    print	$q->Tr(
		       [
			th(
			   [ "Sample Codes","Font Size","&nbsp;","&nbsp;","&nbsp;","&nbsp;"]
			   ),
			td(
			   [
			    $q->popup_menu(-name=>"skeltype",
					   -values=>[sort(keys(%skeltypecode))],
					   -labels=>\%skeltypecode,
					   -defaults=>"",
					   -onChange=>"this.form.source.value=".
					   "makeskelton(this.form) + \"\\n\" + ".
					   "this.form.source.value;"
					   ),
			    $q->popup_menu(-name=>"fontsize",
					   -values=>[sort { $a <=> $b } values(%sourcefsize)],
					   -labels=>\%sourcefsize,
					   -defaults=>$q->param('sourcefontsize'),
					   -onChange=>"document.Rdweb.source.style.fontSize=this.value;".
					   "document.Rdweb.sourcefontsize.value=this.value;".
					   "source_update();"
					   ),
			    $q->button(-name=>'clear', -value=>'Clear',
				       -onClick=>"this.form.source.value=''"),
			    $q->submit(-name=>'logout',
				       -value=>'Logout',
				       -onclick=>'do_action_cancel();'),
			    $q->submit(-name=>'refresh',
				       -value=>'Refresh',
				       -onclick=>'do_action_cancel();'),
			    $q->submit(-name=>'Execute',
				       -value=>'Execute',
				       -onclick=>'do_action_cancel();return(rdweb_check_all(this.form));',
				       ),
			    ]
			   )
			]
		       );
    print $q->end_table();
    
    print
	$q->textarea(
		     -name=>"source",
		     -rows=>20,
		     -columns=>80,
		     -wrap=>"off",
		     -style=>"overflow:scroll;"
		     );
    
    
    print
	$q->end_div;
    
    print
	$q->start_div({-id=>"left"});
    
    # hidden cancel job
    $q->param(-name=>'do_cancel',-value=>'');
    print $q->hidden(-name=>'do_cancel');
    
    print
	$q->start_table ;
    print
	$q->caption(
		    "JOB Information"
		    );
    
    print
	$q->Tr(
	       th([
		   'ID',
		   'Queue',
		   'Status',
		   'Action',
		   ]
		  )
	       );
    
    foreach my $rec (@jobs){
	print $q->start_Tr;
	print $q->td({-type=>'disc'},
		     [
		      ${$rec}{JOBID},
		      ${$rec}{JOBQ},
		      ${$rec}{JOBSTAT},
		      "[".
		      $q->span({-Style=>"cursor: pointer",
				-onClick=>
				    "do_action_clean();".
				    "if(cancel_dialog('".
				    ${$rec}{JOBID}.
				    "',document.Rdweb.do_cancel))".
				    "document.Rdweb.submit();"
				},
			       "Cancel"
			       ).
		      "]",
		      ]
		     );
	print $q->end_Tr;
    }
    print $q->end_table;
    
    # file upload
    print $q->start_table;
    print $q->caption("File Upload");
    print $q->Tr(td(
		    $q->filefield(-class=>'file',
                                  -name=>'uploaded_file',
				  -size=>19,
				  -maxlength=>256)),
		 td(
		    $q->submit(-name=>'upload',-value=>'Upload',
			       -onclick=>'do_action_cancel();'))
		 );
    print $q->end_table;


    # hidden delete file or directory
    $q->param(-name=>'do_deletedir',-value=>'');
    $q->param(-name=>'do_delete',-value=>'');
    print $q->hidden(-name=>'do_deletedir');
    print $q->hidden(-name=>'do_delete');
    
    print $q->start_table;
    print $q->caption("File List");
    print $q->Tr(th({-class=>"dir"},
		    [
		     "[".
		     $q->span({-Style=>"cursor: pointer",
			       -onClick=>"do_action_clean();directory_tree_all_close()"
			       },"+").
		     "/".
		     $q->span({-Style=>"cursor: pointer",
			       -onClick=>"do_action_clean();directory_tree_all_open()"
			       },"-").
		     "]".
		     $rdir,
		     "Action"]
		    ));
    
    foreach my $fileent (sort{
	(((split("/",(split("\t",$b))[4]))[0] cmp (split("/",(split("\t",$a))[4]))[0]) or
	 ((split("\t",$a))[4] cmp (split("\t",$b))[4]) or
	 ((split("\t",$a))[6] cmp (split("\t",$b))[6])) 
	}(@mixdirlist)){
	my ( $type, $size, $level, $datetime, $dirname, $ldirname, $filename ) = split("\t",$fileent);
	my $jsdirname = $dirname;
	$jsdirname =~ s/\//__DIR__/g;
	if (!defined($q->param("dir_xx_$jsdirname"))
	    or $q->param("dir_xx_$jsdirname") eq "" ){
	    $q->param(-name=>"dir_xx_$jsdirname",-value=>'-');
	}




	my $esc = 0;
	my @spdir = split ('/', $dirname);
	my $parent = "";

	for (my $ix = 0; $ix <= $#spdir - (($type eq 'd')?1:0) ; $ix++){
	    
	    $parent .= (($parent eq "")?"":"/");
	    $parent .= $spdir[$ix];
	    my $jsparent = $parent;
	    $jsparent =~ s/\//__DIR__/g;
	    $esc = 1 if ( $q->param("dir_xx_$jsparent") eq "+" );
	}
	if($esc == 1 ){
	    if ( $type eq "d" ){
		print $q->hidden(-name=>"dir_xx_$jsdirname",-default=>'-');
	    }
	}else{
	    print $q->start_Tr();
	    if ( $type eq "d" ){
		print $q->td($q->hidden(-name=>"dir_xx_$jsdirname",-default=>'-'),
			     rep("&nbsp;",$level*3).
			     "[".
			     $q->span({-Style=>"cursor: pointer",
				       -onClick=>
				           "do_action_clean();".
					   "if( document.Rdweb.dir_xx_$jsdirname.value=='+'){".
					   "    document.Rdweb.dir_xx_$jsdirname.value='-';".
					   "}else{".
					   "    document.Rdweb.dir_xx_$jsdirname.value='+';".
					   "}".
					   "document.Rdweb.submit();"},
				      (($q->param("dir_xx_$jsdirname") eq '+')?"+":"-")).
			     "]".
			     $ldirname);
		print $q->td("[".
			     $q->span({-Style=>"cursor: pointer",
				       -onClick=>
				           "do_action_clean();".
					   "if(delete_dir_dialog('$dirname',document.Rdweb.do_deletedir))".
					   "document.Rdweb.submit();"
				       },
				      "Delete"
				      ).
			     "]");
	    }else{
		my $year  = substr($datetime,0,4);
		my $month = substr($datetime,4,2);
		my $day   = substr($datetime,6,2);
		my $HH    = substr($datetime,9,2);
		my $MM    = substr($datetime,11,2);
		my $SS    = substr($datetime,13,2);
		my $filetitle = "$year-$month-$day $HH:$MM:$SS $size bytes"; 

		my $msie =0;
		my $safari =0;
		$msie = 1 if( $q->user_agent() =~ /MSIE/ );
		$safari = 1 if( $q->user_agent() =~ /Safari/ );
		if ($safari){
		    my $script_filename = __FILE__;
		    $script_filename =~ s/^(.*)\/([^\/]*)$/$2/e;
		    my $pathname = $q->url(-absolute=>1);
		    $pathname =~ s/$q->url(-relative=>1)$//;
		    $pathname .= $script_filename . "/" . ((defined($filename))?"$dirname/$filename":$dirname);
		    
		    print $q->td(((defined($filename))?rep("&nbsp;",$level * 3 + 1):rep("&nbsp;", $level * 3)).
				 $q->a({-target=>'_new',
					-title=>$filetitle,
					-href=>$pathname.'?file='. CGI::Util::escape((($dirname ne "")?"$dirname/$filename":$filename)) },
				       ((defined($filename))?$filename:$ldirname)));
		}elsif($msie){
		    print $q->td(((defined($filename))?rep("&nbsp;",$level * 3 + 1):rep("&nbsp;",$level * 3)).
				 $q->a({-target=>'_new',
					-title=>$filetitle,
					-href=>$q->url(-absolute=>1).'?file='. CGI::Util::escape((($dirname ne "")?"$dirname/$filename":$filename)) },
				       ((defined($filename))?$filename:$ldirname)));
		}else{
		    print $q->td(((defined($filename))?rep("&nbsp;",$level * 3 + 1):rep("&nbsp;",$level * 3)).
				 $q->a({-target=>'_new',
					-title=>$filetitle,
					-href=>$q->url(-absolute=>1).'?file='. (($dirname ne "")?"$dirname/$filename":$filename) },
				       ((defined($filename))?$filename:$ldirname)));
		}
		print $q->td("[".
			     $q->span({-Style=>"cursor: pointer",
				       -onClick=>
				           "do_action_clean();".
					   "if(delete_file_dialog('".
					   (($dirname ne "")?"$dirname/$filename":$filename).
					   "',document.Rdweb.do_delete))".
					   "document.Rdweb.submit();"
				       },
				      "Delete"
				      ).
			     "]");
	    }
	    print $q->end_Tr();
	}
    }
    
    print
	$q->end_table;
    
    print
	$q->end_form,
	$q->end_div,
	$q->end_html;
}

sub filedownload($$){
    my $q = shift;
    my $getsid = shift;
    
    my $msie=0;
    $msie=1 if ( $q->user_agent() =~ /MSIE/ );
    my $conv=0;
    
    my $sock = socket_open();
    
    my $user = RemoteUSER($q,$sock,undef,undef,$getsid);
    if( $user eq "" ){
	close($sock);
	RdwebLogin($q);
	return;
    }
    my $home = RemotePWD($sock);
    my $rdir = RemoteMKDIR($sock,"$home/$Rdweb::dir");
    $rdir = RemoteCD($sock,"$home/$Rdweb::dir");
    
    my $fnm = $rdir . "/" . $q->param('file');

    my $escname = $q->param('file');
    $escname =~ s/^(.)+\///;
    if ($msie==1){
	$escname =~ s/([^\._0-9A-Za-z])/"%".unpack("H*",$1)/eg;
    }else{
	$escname =~ s/([^\._0-9A-Za-z])/"%".unpack("H*",$1)/eg;
	$escname = "utf-8'en'$escname" ;
    }

    my %mime=();
    
    if( $fnm =~ /\.ps$/ ||  $fnm =~ /\.eps$/ ) {
	$mime{'type'}='Application/Octet-Stream';
	$mime{'expires'}='now';
	$mime{'attachment'}=1;
    }elsif ( $fnm =~ /\.pdf$/ ){
	$mime{'type'}='Application/Pdf';
	$mime{'expires'}='now';
	$mime{'attachment'}=1;
    }elsif ( $fnm =~ /\.sh$/ ){
	$mime{'type'}='text/plain';
	$mime{'charset'}='utf-8';
	$mime{'expires'}='now';
    }elsif ( $fnm =~ /\.Rout$/ ){
	$mime{'type'}='text/plain';
	$mime{'charset'}='utf-8';
	$mime{'expires'}='now';
    }elsif ( $fnm =~ /\.log$/ ){
	$mime{'type'}='text/plain';
	$mime{'charset'}='utf-8';
	$mime{'expires'}='now';
    }elsif ( $fnm =~ /\.out$/ ){
	$mime{'type'}='text/plain';
	$mime{'charset'}='utf-8';
	$mime{'expires'}='now';
    }elsif ( $fnm =~ /\.err$/ ){
	$mime{'type'}='text/plain';
	$mime{'charset'}='utf-8';
	$mime{'expires'}='now';
    }elsif ( $fnm =~ /\.txt$/ ){
	$mime{'type'}='text/plain';
	$mime{'charset'}='utf-8';
	$mime{'expires'}='now';
    }elsif ( $fnm =~ /\.R$/ ){
	if($msie==1){
	    $mime{'type'}='text/plain';
	    $mime{'charset'}='utf-8';
	    $mime{'expires'}='now';
	    $mime{'attachment'}=1;
	    $conv=1;
	}else{
	    $mime{'type'}='text/plain';
	    $mime{'charset'}='utf-8';
	    $mime{'expires'}='now';
	}
    }elsif ( $fnm =~ /\.png$/ ){
	$mime{'type'}='image/png';
	$mime{'expires'}='now';
	$mime{'attachment'}=1;
    }elsif ( $fnm =~ /\.jpg$/ || $fnm =~ /\.jpeg$/  ){
	$mime{'type'}='image/jpg';
	$mime{'expires'}='now';
	$mime{'attachment'}=1;
    }else{
	$mime{'type'}='Application/Octet-Stream';
	$mime{'charset'}='utf-8';
	$mime{'expires'}='now';
	$mime{'attachment'}=1;
    }

    if (defined($mime{'attachment'})){
	if ($msie==1){
	    $mime{'attachment'}=$escname;
	}else{
	    delete $mime{'attachment'};
	    $mime{'Content_Disposition'}="attachment; filename*=$escname";
	}
    }

    print
	$q->header(\%mime);

# -Content_Disposition => "attachment; filename*=$escname");

    open(FD, ">>/tmp/log.txt");
    print FD "PATH_INFO", $ENV{PATH_INFO}, "\n";
    print FD "escname", $escname, "\n";
    close(FD);

    print $sock "load $fnm".$EOL;

    if ( $msie == 1 and $conv == 1){  # LF => CRLF 
	my $text="";
	while(my $buf=<$sock>){
	    last if ( $buf =~ /^EOF/ );
	    $buf =~ s/[\r\n]//g;
	    $text .= pack("H*",$buf);
	}
	$text =~ s/\n/\r\n/g;
	print $text;
    }else{
	while(my $buf=<$sock>){
	    last if ( $buf =~ /^EOF/ );
	    $buf =~ s/[\r\n]//g;
	    print pack("H*",$buf);
	}
    }

    close ($sock);
}

sub fileupload($$){
    my $q = shift;
    my $getsid = shift;
    
    my $sock = socket_open();
    
    my $user = RemoteUSER($q,$sock,undef,undef,$getsid);
    if( $user eq "" ){
	close($sock);
	RdwebLogin($q);
	return;
    }
    
    my $home = RemotePWD($sock);
    my $rdir = RemoteMKDIR($sock,"$home/$Rdweb::dir");
    $rdir = RemoteCD($sock,"$home/$Rdweb::dir");
    
    my $fnm = $q->param('uploaded_file');
    
    my $filename = $fnm;
    
    $filename =~ s/^.:.*\\//g;
    #$filename =~ s/([^0-9A-Za-z_+\.])/unpack("H*",$1)/eg;
    
    
    open(FD, ">>/tmp/log.txt");
    print FD "param=", join(':',$q->param()) ,"\n";
    print FD "fnm=", $fnm ,"\n";
    print FD "filename=", $filename ,"\n";
    close(FD);
    
    print $sock "save $filename".$EOL;
    my $buf;
    my $length;
    while($length = sysread($fnm,$buf,4096)){
	last unless ($length > 0);
	print $sock unpack("H*",$buf), $EOL;
    }
    print $sock "EOF".$EOL;
    <$sock>; # EOF
    close ($sock);
}

sub dirdelete($$){
    my $q = shift;
    my $getsid = shift;
    
    my $sock = socket_open();
    
    my $user = RemoteUSER($q,$sock,undef,undef,$getsid);
    if( $user eq "" ){
	close($sock);
	RdwebLogin($q);
	return;
    }
    
    my $home = RemotePWD($sock);
    my $rdir = RemoteMKDIR($sock,"$home/$Rdweb::dir");
    $rdir = RemoteCD($sock,"$home/$Rdweb::dir");
    
    my $fnm = $rdir . "/" . $q->param('do_deletedir');
    
    print $sock "rmdir $fnm".$EOL;
    my $buf = <$sock>;
    unless ( $buf =~ /^RMDIR=/ ){
	close($sock);
	die "failed delete directory $fnm.";
    }
    close ($sock);
}

sub filedelete($$){
    my $q = shift;
    my $getsid = shift;
    
    my $sock = socket_open();
    
    my $user = RemoteUSER($q,$sock,undef,undef,$getsid);
    if( $user eq "" ){
	close($sock);
	RdwebLogin($q);
	return;
    }
    my $home = RemotePWD($sock);
    my $rdir = RemoteMKDIR($sock,"$home/$Rdweb::dir");
    $rdir = RemoteCD($sock,"$home/$Rdweb::dir");
    
    my $fnm = $rdir . "/" . $q->param('do_delete');
    
    print $sock "unlink $fnm".$EOL;
    my $buf = <$sock>;
    unless ( $buf =~ /^UNLINK=/ ){
	close($sock);
	die "failed delete $fnm.";
    }
    close ($sock);
}

sub RdwebLogin($){
    my $q = shift;
    my $fqdn = $q->url(-full=>1);
    my $secure = 0;
    my $cookiepath = $q->url(-absolute=>1);
    $cookiepath =~ s/[^\/]+$//g;
    if ( $fqdn =~ /https:\/\// ){
	$secure=1;
    }
    $fqdn =~ s/^([0-9A-Za-z]+):\/\///;

    my @save_params = ();
    foreach my $pname ($q->param())
    {
	push @save_params, $pname if ( $pname =~ /^(useQueue|useSlave|useBLAS|cltype)$/ );
	push @save_params, $pname if ( $pname =~ /^source/ );
	push @save_params, $pname if ( $pname =~ /^dir_xx_/ );
    }


    print $q->header(
		     -cookie=>$q->cookie(-name=>'SID',
					 -value=>'',
					 -expires=>'-24h',
					 -path=>$cookiepath,
					 -host=>$fqdn,
					 -secure=>$secure
					 ),
		     -expires=>"now",
		     -charset=>'utf-8');
    print $q->start_html(-title=>'RdwebLogin',
			 -style=>{'src'=>"index.${Rdweb::host}.css"},
			 -lang=>'en_US',
			 -encoding=>'utf-8');
    
    print $q->h1("Let's Use Parallel <img height=36 src=\"images/Rlogo-5.gif\" alt=\"R\" />.$H1hostname");
    print $COPYRIGHT;
    print $q->start_multipart_form(-name=>"Rdweb");
    print $q->start_div({-id=>"main",-align=>"center"});
    print $q->start_table;
    print $q->caption('Please input the information of the account on the super computer');
    print $q->Tr($q->th('Username'),$q->td({-align=>'left'},$q->textfield('Rdwebue')));
    print $q->Tr($q->th('Password'),$q->td({-align=>'left'},$q->password_field('Rdwebpd')));
    print $q->end_table;
    print $q->submit(-name=>'login',-value=>'Login'),$q->button(-name=>'cancel',-value=>'Cancel',-onClick=>"this.form.Rdwebue.value='';this.form.Rdwebpd.value=''");
    print $q->end_div;
    foreach my $pname (@save_params){
	print $q->hidden(-name=>$pname,-value=>$q->param($pname));
    }
    print $q->end_form();
    print $q->end_html;  
}

sub check_grant($){
    my $q = shift;
    my $sock = socket_open();
    my $sid = RemoteUSER($q,$sock,$q->param('Rdwebue'),$q->param('Rdwebpd'),undef);
    close($sock);
    return $sid;
}


#=============================================================================
package main;


my $q = new CGI;
my $getsid = $q->cookie('SID');

if(defined($q->param('login')) &&  $q->param('login') ne ""){
    Rdweb($q,undef);
    exit();
}

if ( defined($q->param('logout')) && $q->param('logout') eq "Logout"){
    RdwebLogin($q);
    exit(0);
}

if ( defined($q->param('refresh')) && $q->param('refresh') eq "Refresh"){
    Rdweb($q,$getsid);
    exit(0);
}

if ( defined($q->param('Execute')) && $q->param('Execute') eq "Execute"){
    submitJOB($q,$getsid);
    Rdweb($q,$getsid);
    exit(0);
}

if(defined($q->param('file'))){
    filedownload($q,$getsid);
    exit();
}
if ( defined($q->param('upload')) && $q->param('upload') ne ""){
    open(FD, ">>/tmp/log.txt");
    print FD $q->param('upload'), "start\n";
    close(FD);

    fileupload($q,$getsid);

    open(FD, ">>/tmp/log.txt");
    print FD $q->param('upload'), "end\n";
    close(FD);
    Rdweb($q,$getsid);
    exit(0);
}
if ( defined($q->param('do_deletedir')) && $q->param('do_deletedir') ne ""){
    dirdelete($q,$getsid);
    Rdweb($q,$getsid);
    exit(0);
}
if ( defined($q->param('do_delete')) && $q->param('do_delete') ne ""){
    filedelete($q,$getsid);
    Rdweb($q,$getsid);
    exit(0);
}
if ( defined($q->param('do_cancel')) &&  $q->param('do_cancel') ne ""){
    cancelJOB($q,$getsid);
    Rdweb($q,$getsid);
    exit(0);
}
Rdweb($q,$getsid);
exit(0);

